/******************************************************************************
 * Copyright (c) Huawei Technologies Co., Ltd. 2021. All rights reserved.
 * gala-gopher licensed under the Mulan PSL v2.
 * You can use this software according to the terms and conditions of the Mulan PSL v2.
 * You may obtain a copy of Mulan PSL v2 at:
 *     http://license.coscl.org.cn/MulanPSL2
 * THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND, EITHER EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT, MERCHANTABILITY OR FIT FOR A PARTICULAR
 * PURPOSE.
 * See the Mulan PSL v2 for more details.
 * Author: luzhihao
 * Create: 2022-05-16
 * Description:
 ******************************************************************************/
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <stdbool.h>
#include <unistd.h>
#include <time.h>
#include <stdarg.h>
#include "common.h"
#include "event_config.h"
#include "event.h"
#ifdef NATIVE_PROBE_FPRINTF
#include "nprobe_fprintf.h"
#endif

static struct evt_ts_hash_t *g_evt_head = NULL;
static unsigned int g_evt_period = 600;
static EventsConfig *g_evt_conf;
static char g_lang_type[MAX_EVT_GRP_NAME_LEN] = "zh_CN";

static void hash_clear_older_evt(time_t cur_time);
static unsigned int hash_count_evt(void);
static int is_evt_need_report(const char *entityId, time_t cur_time);

static void __get_local_time(char *buf, int buf_len, time_t *cur_time)
{
    time_t rawtime;
    struct tm tm;
    char time_str[TIME_STRING_LEN];

    (void)time(&rawtime);
    asctime_r(localtime_r(&rawtime, &tm), time_str);
    SPLIT_NEWLINE_SYMBOL(time_str);
    (void)snprintf(buf, (const int)buf_len, "%s", time_str);
    *cur_time = rawtime;
}

static void __replace_desc_fmt(const char* entityName, const char* metrics, const char *fmt, char *new_fmt)
{
    if (new_fmt == NULL) {
        return;
    }
    (void)strncpy(new_fmt, fmt, MAX_EVT_BODY_LEN - 1);

    if (g_lang_type[0] == 0) {
        return;
    }
    if (g_evt_conf == NULL) {
        if (events_config_init(&g_evt_conf, g_lang_type) < 0) {
            ERROR("event report failed, bacause language type(%s) but config_init failed.\n", g_lang_type);
            return;
        }
    }
    if (get_event_field(g_evt_conf, entityName, metrics, new_fmt) == 0) {
        ERROR("get evt entity[%s] metric[%s] field failed.\n", entityName, metrics);
        return;
    }
    return;
}

#define __SEC_TXT_LEN  32
struct evt_sec_s {
    int sec_number;
    char sec_text[__SEC_TXT_LEN];
};

static struct evt_sec_s secs[EVT_SEC_MAX] = {
    {9,              "INFO"},
    {13,              "WARN"},
    {17,              "ERROR"},
    {21,              "FATAL"}
};

#define __EVT_BODY_LEN  512	// same as MAX_IMDB_METRIC_VAL_LEN
void report_logs(const char* entityName,
                 const char* entityId,
                 const char* metrics,
                 enum evt_sec_e sec,
                 const char * fmt, ...)
{
    int len;
    va_list args;
    char body[__EVT_BODY_LEN];
    char *p;
    time_t cur_time;

    body[0] = 0;
    __get_local_time(body, __EVT_BODY_LEN, &cur_time);
    if ((g_evt_period > 0) && (!is_evt_need_report(entityId, cur_time))) {
        DEBUG("event not report, bacause entityId[%s] in event_period.\n", entityId);
        return;
    }

    p = body + strlen(body);
    len = __EVT_BODY_LEN - strlen(body);

    (void)snprintf(p, len, " %s Entity(%s) ", secs[sec].sec_text, entityId);
    p = body + strlen(body);
    len = __EVT_BODY_LEN - strlen(body);

    char fmt2[MAX_EVT_BODY_LEN];
    va_start(args, fmt);
    __replace_desc_fmt(entityName, metrics, fmt, fmt2);
    (void)vsnprintf(p, len, fmt2, args);
    va_end(args);

#ifdef NATIVE_PROBE_FPRINTF
    (void)nprobe_fprintf(stdout, "|%s|%s|%s|%s|%s|%d|%s|\n",
                         "event",
                         entityName,
                         entityId,
                         metrics,
                         secs[sec].sec_text,
                         secs[sec].sec_number,
                         body);
#else
    (void)fprintf(stdout, "|%s|%s|%s|%s|%s|%d|%s|\n",
                          "event",
                          entityName,
                          entityId,
                          metrics,
                          secs[sec].sec_text,
                          secs[sec].sec_number,
                          body);
#endif
    return;
}

static void hash_add_evt(const char *entityId, time_t cur_time)
{
    struct evt_ts_hash_t *item = NULL;

    if (hash_count_evt() >= MAX_EVT_NUM) {
        // clear older events when event num beyond 1000
        hash_clear_older_evt(cur_time);
    }

    item = malloc(sizeof(struct evt_ts_hash_t));
    if (item == NULL) {
        ERROR("event malloc error\n");
        return;
    }
    item->entity_id[0] = 0;
    strncpy(item->entity_id, entityId, MAX_ENTITY_NAME_LEN - 1);
    item->evt_ts = cur_time;
    H_ADD_S(g_evt_head, entity_id, item);
}

static struct evt_ts_hash_t *hash_find_evt(const char *entityId)
{
    char str[MAX_ENTITY_NAME_LEN];
    struct evt_ts_hash_t *item = NULL;
    if (g_evt_head == NULL) {
        return NULL;
    }

    str[0] = 0;
    strncpy(str, entityId, MAX_ENTITY_NAME_LEN - 1);
    H_FIND_S(g_evt_head, str, item);
    if (item == NULL) {
        return NULL;
    }
    return item;
}

static void hash_clear_older_evt(time_t cur_time)
{
    if (g_evt_head == NULL) {
        return;
    }

    struct evt_ts_hash_t *item, *tmp;
    H_ITER(g_evt_head, item, tmp) {
        if ((cur_time - item->evt_ts) <= g_evt_period) {
            continue;
        }
        H_DEL(g_evt_head, item);
        (void)free(item);
    }
}

static unsigned int hash_count_evt(void)
{
    unsigned int evt_num;
    evt_num = H_COUNT(g_evt_head);
    return evt_num;
}

static int is_evt_need_report(const char *entityId, time_t cur_time)
{
    struct evt_ts_hash_t *item = NULL;

    item = hash_find_evt(entityId);
    if (item == NULL) {
        hash_add_evt(entityId, cur_time);
        if (hash_find_evt(entityId) == NULL) {
            ERROR("evt_hash add eventid[%s] failed.\n", entityId);
            return 0;
        } else {
            return 1;
        }
    }
    if ((cur_time > item->evt_ts) && (cur_time - item->evt_ts >= g_evt_period)) {
        item->evt_ts = cur_time;
        return 1;
    }
    return 0;
}

void init_event_mgr(unsigned int time_out, char *lang_type)
{
    g_evt_period = time_out;
    g_lang_type[0] = 0;
    if (lang_type != NULL && strlen(lang_type) > 0) {
        (void)strncpy(g_lang_type, lang_type, MAX_EVT_GRP_NAME_LEN - 1);
    }
}
